from pathlib import Path
from tempfile import TemporaryDirectory, NamedTemporaryFile
import os
import unittest

from dhpython.tools import fix_shebang, relpath, move_matching_files


class TestRelpath(unittest.TestCase):
    def test_common_parent_dir(self) -> None:
        r = relpath("/usr/share/python-foo/foo.py", "/usr/bin/foo")
        self.assertEqual(r, "../share/python-foo/foo.py")

    def test_strips_common_prefix(self) -> None:
        r = relpath("/usr/share/python-foo/foo.py", "/usr/share")
        self.assertEqual(r, "python-foo/foo.py")

    def test_trailing_slash_ignored(self) -> None:
        r = relpath("/usr/share/python-foo/foo.py", "/usr/share/")
        self.assertEqual(r, "python-foo/foo.py")


class TestMoveMatchingFiles(unittest.TestCase):
    def setUp(self) -> None:
        self.tmpdir = TemporaryDirectory()  # pylint: disable=consider-using-with
        self.addCleanup(self.tmpdir.cleanup)
        os.makedirs(self.tmppath("foo/bar/a/b/c/spam"))
        for path in ("foo/bar/a/b/c/spam/file.so", "foo/bar/a/b/c/spam/file.py"):
            with open(self.tmppath(path), "wb"):
                # create a 0 byte file for the test
                pass

        move_matching_files(
            self.tmppath("foo/bar/"), self.tmppath("foo/baz/"), r"spam/.*\.so$"
        )

    def tmppath(self, *path: str) -> str:
        return os.path.join(self.tmpdir.name, *path)

    def test_moved_matching_file(self) -> None:
        self.assertTrue(os.path.exists(self.tmppath("foo/baz/a/b/c/spam/file.so")))

    def test_left_non_matching_file(self) -> None:
        self.assertTrue(os.path.exists(self.tmppath("foo/bar/a/b/c/spam/file.py")))


class TestFixShebang(unittest.TestCase):
    def setUp(self) -> None:
        self.tmpfile = Path(
            NamedTemporaryFile(
                prefix="dhptest_", suffix="_shebang.py", delete=False
            ).name
        )
        self.addCleanup(self.tmpfile.unlink)

    def write_shebang(self, shebang: str) -> None:
        self.tmpfile.write_text(shebang + "\nprint('This is Python')\n")

    def assert_shebang(self, shebang: str) -> None:
        contents = self.tmpfile.read_text().splitlines()
        self.assertEqual(len(contents), 2)
        self.assertEqual(contents[0], shebang)
        self.assertEqual(contents[1], "print('This is Python')")

    def test_perl(self) -> None:
        self.write_shebang("#!/usr/bin/perl")
        fix_shebang(self.tmpfile)
        self.assert_shebang("#!/usr/bin/perl")

    def test_unversioned(self) -> None:
        self.write_shebang("#!/usr/bin/python")
        fix_shebang(self.tmpfile)
        self.assert_shebang("#! /usr/bin/python3")

    def test_python2(self) -> None:
        self.write_shebang("#!/usr/bin/python2")
        fix_shebang(self.tmpfile)
        self.assert_shebang("#!/usr/bin/python2")

    def test_python2_7(self) -> None:
        self.write_shebang("#!/usr/bin/python2.7")
        fix_shebang(self.tmpfile)
        self.assert_shebang("#!/usr/bin/python2.7")

    def test_python3(self) -> None:
        self.write_shebang("#!/usr/bin/python3")
        fix_shebang(self.tmpfile)
        self.assert_shebang("#! /usr/bin/python3")

    def test_python3_13(self) -> None:
        self.write_shebang("#!/usr/bin/python3.13")
        fix_shebang(self.tmpfile)
        self.assert_shebang("#!/usr/bin/python3.13")

    def test_env_unversioned(self) -> None:
        self.write_shebang("#!/usr/bin/env python")
        fix_shebang(self.tmpfile)
        self.assert_shebang("#! /usr/bin/python3")

    def test_env_python3(self) -> None:
        self.write_shebang("#!/usr/bin/env python3")
        fix_shebang(self.tmpfile)
        self.assert_shebang("#! /usr/bin/python3")

    def test_env_python3_13(self) -> None:
        self.write_shebang("#!/usr/bin/env python3.13")
        fix_shebang(self.tmpfile)
        self.assert_shebang("#! /usr/bin/python3.13")

    def test_replacement(self) -> None:
        self.write_shebang("#!/usr/bin/env python")
        fix_shebang(self.tmpfile, "/usr/bin/foo")
        self.assert_shebang("#! /usr/bin/foo")



#! /usr/bin/python3

import DistUtilsExtra.auto

DistUtilsExtra.auto.setup(name='Foo',
      version='0.1',
      description="package with private Python extension",
      author='Maximilian Engelhardt',
      author_email='maxi@daemonizer.de',
      url='http://www.debian.org/',
)



#! /usr/bin/python3
# Copyright © 2010-2015 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import re
import sys

try:
    from distro_info import DistroInfo  # python3-distro-info package
except ImportError:
    DistroInfo = None
from gzip import decompress
from os import chdir, mkdir
from os.path import dirname, exists, isdir, join, split
from urllib.request import urlopen

if "--ubuntu" in sys.argv and DistroInfo:
    SOURCES = [
        "http://archive.ubuntu.com/ubuntu/dists/%s/Contents-amd64.gz"
        % DistroInfo("ubuntu").devel(),
    ]
else:
    SOURCES = [
        "http://ftp.debian.org/debian/dists/unstable/main/Contents-all.gz",
        "http://ftp.debian.org/debian/dists/unstable/main/Contents-amd64.gz",
    ]

IGNORED_PKGS = set()
OVERRIDES = {
    "cpython3": {
        "argparse": "python3 (>= 3.2)",
        "cython": "cython3",
        "pil": "python3-pil",
        "pillow": "python3-pil",
        "pylint": "pylint",
        "pyside6": None,
    },
}

public_egg = re.compile(
    r"""
    /usr/
    (
        (?P<cpython3>
            (lib/python3/dist-packages)
        )
    )
    /[^/]*\.(dist|egg)-info
""",
    re.VERBOSE,
).match

skip_sensible_names = True if "--skip-sensible-names" in sys.argv else False

chdir(dirname(__file__))
if isdir("../dhpython"):
    sys.path.append("..")
else:
    sys.path.append("/usr/share/dh-python/dhpython/")
from dhpython.pydist import normalize_name, sensible_pname

data = ""
if not isdir("cache"):
    mkdir("cache")
for source in SOURCES:
    cache_fpath = join("cache", split(source)[-1])
    if not exists(cache_fpath):
        with urlopen(source) as fp:
            source_data = fp.read()
        with open(cache_fpath, "wb") as fp:
            fp.write(source_data)
    else:
        with open(cache_fpath, "rb") as fp:
            source_data = fp.read()
    try:
        data += str(decompress(source_data), encoding="UTF-8")
    except UnicodeDecodeError as e:  # Ubuntu
        data += str(decompress(source_data), encoding="ISO-8859-15")

result = {"cpython3": {}}

for line in data.splitlines():
    try:
        path, desc = line.rsplit(maxsplit=1)
    except ValueError:
        # NOTE(jamespage) some lines in Ubuntu are not parseable.
        continue
    path = "/" + path.rstrip()
    section, pkg_name = desc.rsplit("/", 1)
    if pkg_name in IGNORED_PKGS:
        continue
    match = public_egg(path)
    if match:
        dist_name = [
            i.split("-", 1)[0]
            for i in path.split("/")
            if i.endswith((".egg-info", ".dist-info"))
        ][0]
        if dist_name.endswith(".egg"):
            dist_name = dist_name[:-4]
        dist_name = normalize_name(dist_name)

        impl = next(key for key, value in match.groupdict().items() if value)

        if skip_sensible_names and sensible_pname(impl, dist_name) == pkg_name:
            continue

        if impl not in result:
            continue

        processed = result[impl]
        if dist_name not in processed:
            processed[dist_name] = pkg_name

for impl, details in result.items():
    with open(f"{impl}_fallback", "w") as fp:
        overrides = OVERRIDES[impl]
        lines = []
        for egg, value in overrides.items():
            if value:
                lines.append(f"{egg} {value}\n")
        lines.extend(
            f"{egg} {pkg}\n" for egg, pkg in details.items() if egg not in overrides
        )
        fp.writelines(sorted(lines))



# Copyright © 2012-2013 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging
from collections.abc import Callable
from functools import wraps
from glob import glob1
from os import remove
from os.path import exists, isdir, join
from shutil import rmtree, move
from typing import (
    TYPE_CHECKING,
    Concatenate,
    Literal,
    ParamSpec,
    TypeVar,
    override
)

from dhpython.build.base import Base, shell_command, copy_test_files, create_setuptools_parallel_cfg
from dhpython.build.types import Args, Context

if TYPE_CHECKING:
    from functools import _Wrapped


log = logging.getLogger('dhpython')
_setup_tpl = 'setup.py|setup-3.py'


P = ParamSpec('P')
T = TypeVar('T')

def create_pydistutils_cfg(
    func: Callable[Concatenate["BuildSystem", Context, Args, P], T]
) -> "_Wrapped[Concatenate[BuildSystem, Context, Args, P], T, Concatenate[BuildSystem, Context, Args, P], T]":
    """distutils doesn't have sane command-line API - this decorator creates
    .pydistutils.cfg file to workaround it

    hint: if you think this is plain stupid, please don't read
    distutils/setuptools/distribute sources
    """

    @wraps(func)
    def wrapped_func(
        self: "BuildSystem",
        context: Context,
        args: Args,
        *oargs: P.args,
        **kwargs: P.kwargs
    ) -> T:
        fpath = join(args.home_dir, '.pydistutils.cfg')
        if not exists(fpath):
            with open(fpath, 'w', encoding='utf-8') as fp:
                lines = ['[clean]\n',
                         'all=1\n',
                         '[build]\n',
                         'build_lib={}\n'.format(args.build_dir),
                         '[install]\n',
                         'force=1\n',
                         'install_layout=deb\n',
                         'install_scripts=$base/bin\n',
                         'install_lib={}\n'.format(args.install_dir),
                         'prefix=/usr\n']
                log.debug('pydistutils config file:\n%s', ''.join(lines))
                fp.writelines(lines)
        context.ENV['HOME'] = args.home_dir
        return func(self, context, args, *oargs, **kwargs)

    return wrapped_func


class BuildSystem(Base):
    DESCRIPTION = 'Distutils build system'
    SUPPORTED_INTERPRETERS = {'python', 'python3', 'python{version}',
                              'python-dbg', 'python3-dbg', 'python{version}-dbg'}
    REQUIRED_FILES = [_setup_tpl]
    OPTIONAL_FILES = {'setup.cfg': 1,
                      'requirements.txt': 1,
                      'PKG-INFO': 10,
                      '*.egg-info': 10}
    CLEAN_FILES = Base.CLEAN_FILES | {'build'}

    @override
    def detect(self, context: Context) -> int:
        result = super().detect(context)
        if _setup_tpl in self.DETECTED_REQUIRED_FILES:
            context.args['setup_py'] = self.DETECTED_REQUIRED_FILES[_setup_tpl][0]
        else:
            context.args['setup_py'] = 'setup.py'
        return result

    @shell_command
    @create_pydistutils_cfg
    @override
    def clean(self, context: Context, args: Args) -> str | Literal[0]:
        super().clean(context, args)
        if exists(args.interpreter.binary()):
            return '{interpreter} {setup_py} clean {args}'
        return 0  # no need to invoke anything

    @shell_command
    @create_pydistutils_cfg
    @override
    def configure(self, context: Context, args: Args) -> str:
        return '{interpreter} {setup_py} config {args}'

    @shell_command
    @create_pydistutils_cfg
    @override
    def build(self, context: Context, args: Args) -> str:
        return '{interpreter.binary_dv} {setup_py} build {args}'

    @shell_command
    @create_setuptools_parallel_cfg()
    def _bdist_wheel(self, context: Context, args: Args) -> str:
        # pylint: disable=unused-argument
        try:
            # pylint: disable=unused-import
            import wheel  # type: ignore
        except ImportError:
            raise Exception("wheel is required to build wheels for distutils/setuptools packages. Build-Depend on python3-wheel.")
        # Don't build a wheel in the deb install layout
        fpath = join(args.home_dir, '.pydistutils.cfg')
        remove(fpath)
        return '{interpreter.binary_dv} -c "import setuptools, runpy; runpy.run_path(\'{setup_py}\')" bdist_wheel {args}'

    @override
    def build_wheel(self, context: Context, args: Args) -> None:
        self._bdist_wheel(context, args)
        dist_dir = join(args.dir, 'dist')
        wheels = glob1(dist_dir, '*.whl')
        n_wheels = len(wheels)
        if n_wheels != 1:
            raise Exception(f"Expected 1 wheel, found {n_wheels}")
        move(join(dist_dir, wheels[0]), args.home_dir)

    @shell_command
    @create_pydistutils_cfg
    @override
    def install(self, context: Context, args: Args) -> str:
        # remove egg-info dirs from build_dir
        for fname in glob1(args.build_dir, '*.egg-info'):
            fpath = join(args.build_dir, fname)
            if isdir(fpath):
                rmtree(fpath)
            else:
                remove(fpath)

        return '{interpreter.binary_dv} {setup_py} install --root {destdir} {args}'

    @shell_command
    @create_pydistutils_cfg
    @copy_test_files()
    @override
    def test(self, context: Context, args: Args) -> str:
        return super().test_cmd(context, args)



from unittest import TestCase


class RequiredTest(TestCase):
    def test_tests_are_executed(self):
        open('test-executed', 'w').close()

    def test_testfiles_exist(self):
        open('testfile1.txt').close()
        open('testfile2.txt').close()
        open('testdir/testfile3.txt').close()
        open('destination/testfile4.txt').close()
        open('renamed').close()



import os
import unittest
from tempfile import TemporaryDirectory
from typing import Any

from dhpython.debhelper import DebHelper, build_options, Options


class DebHelperTestCase(unittest.TestCase):
    impl = "cpython3"
    control: list[str] = []
    options: dict[str, Any] = {}
    compat: int | None = None
    parse_control = True

    def build_options(self) -> Options:
        return build_options(**self.options)

    def setUp(self) -> None:
        self.tempdir = TemporaryDirectory()  # pylint: disable=consider-using-with
        self.addCleanup(self.tempdir.cleanup)

        old_wd = os.getcwd()
        os.chdir(self.tempdir.name)
        self.addCleanup(os.chdir, old_wd)

        os.mkdir("debian")
        with open("debian/control", "w", encoding="UTF-8") as f:
            f.write("\n".join(self.control))
        if self.compat:
            with open("debian/compat", "w", encoding="UTF-8") as f:
                f.write(f"{self.compat}\n")
        if self.parse_control:
            self.dh = DebHelper(self.build_options(), impl=self.impl)


CONTROL = [
    "Source: foo-src",
    "Build-Depends: debhelper-compat (= 13), ",
    " python3-all, ",
    " python-all",
    "\t, bar (<< 2) [amd64],",
    " baz (>= 1.0)",
    "X-Python3-Version: >= 3.1, << 3.10",
    "",
    "Architecture: all",
    "Package: python3-foo",
    "Depends: ${python3:Depends}",
    "",
    "Package: python3-foo-ext",
    "Architecture: any",
    "Depends: ${python3:Depends}, " "# COMMENT",
    " ${shlibs:Depends},",
    "",
    "Package: python-foo",
    "Architecture: all",
    "Depends: ${python:Depends}",
    "",
    "",
    "Package: foo",
    "Architecture: all",
    "Depends: ${python3:Depends}",
    "",
    "",
    "Package: recfoo",
    "Architecture: all",
    "Recommends: ${python3:Depends}",
    "",
    "",
]


class TestControlBlockParsing(DebHelperTestCase):
    control = CONTROL

    def test_parses_source(self) -> None:
        self.assertEqual(self.dh.source_name, "foo-src")

    def test_parses_build_depends(self) -> None:
        self.assertEqual(
            self.dh.build_depends,
            {
                "debhelper-compat": {None: "= 13"},
                "python3-all": {None: None},
                "python-all": {None: None},
                "bar": {"amd64": "<< 2"},
                "baz": {None: ">= 1.0"},
            },
        )

    def test_parses_XPV(self) -> None:
        self.assertEqual(self.dh.python_version, ">= 3.1, << 3.10")

    def test_parses_packages(self) -> None:
        self.assertEqual(
            list(self.dh.packages.keys()),
            ["python3-foo", "python3-foo-ext", "foo", "recfoo"],
        )

    def test_parses_arch(self) -> None:
        self.assertEqual(self.dh.packages["python3-foo-ext"].arch, "any")

    def test_parses_arch_all(self) -> None:
        self.assertEqual(self.dh.packages["python3-foo"].arch, "all")

    def test_compat(self) -> None:
        self.assertEqual(self.dh.compat(), 13)


class TestControlSkipIndep(DebHelperTestCase):
    control = CONTROL
    options = {
        "arch": True,
    }

    def test_skip_indep(self) -> None:
        self.assertEqual(list(self.dh.packages.keys()), ["python3-foo-ext"])


class TestControlSkipArch(DebHelperTestCase):
    control = CONTROL
    options = {
        "arch": False,
    }

    def test_skip_arch(self) -> None:
        self.assertEqual(
            list(self.dh.packages.keys()), ["python3-foo", "foo", "recfoo"]
        )


class TestControlSinglePkg(DebHelperTestCase):
    control = CONTROL
    options = {
        "package": ["python3-foo"],
    }

    def test_parses_packages(self) -> None:
        self.assertEqual(list(self.dh.packages.keys()), ["python3-foo"])


class TestControlSkipSinglePkg(DebHelperTestCase):
    control = CONTROL
    options = {
        "no_package": ["python3-foo"],
    }

    def test_parses_packages(self) -> None:
        self.assertEqual(
            list(self.dh.packages.keys()), ["python3-foo-ext", "foo", "recfoo"]
        )


class TestControlNoBinaryPackages(DebHelperTestCase):
    control = [
        "Source: foo-src",
        "Build-Depends: python3-all",
        "",
    ]
    parse_control = False

    def test_throws_error(self) -> None:
        msg = "Unable to parse debian/control, found less than 2 paragraphs"
        with self.assertRaisesRegex(Exception, msg):
            DebHelper(self.build_options())


class TestControlMissingPackage(DebHelperTestCase):
    control = [
        "Source: foo-src",
        "Build-Depends: python3-all",
        "",
        "Architecture: all",
    ]
    parse_control = False

    def test_parses_packages(self) -> None:
        msg = "Unable to parse debian/control, paragraph 2 missing Package " "field"
        with self.assertRaisesRegex(Exception, msg):
            DebHelper(self.build_options())


class TestRemainingPackages(DebHelperTestCase):
    control = CONTROL
    options = {
        "remaining_packages": True,
    }
    parse_control = False

    def setUp(self) -> None:
        super().setUp()
        with open("debian/python3-foo.debhelper.log", "w", encoding="UTF-8") as f:
            f.write("dh_python3\n")
        with open("debian/python3-foo-ext.debhelper.log", "w", encoding="UTF-") as f:
            f.write("dh_foobar\n")
        self.dh = DebHelper(self.build_options(), impl=self.impl)

    def test_skips_logged_packages(self) -> None:
        self.assertEqual(
            list(self.dh.packages.keys()), ["python3-foo-ext", "foo", "recfoo"]
        )


class TestLegacyCompatFile(DebHelperTestCase):
    control = [
        "Source: foo-src",
        "Build-Depends: python3-all",
        "",
        "Package: foo",
        "Architecture: all",
    ]
    compat = 10
    parse_control = True

    def test_compat(self) -> None:
        self.assertEqual(self.dh.compat(), 10)


class TestNoCompatLevel(DebHelperTestCase):
    control = [
        "Source: foo-src",
        "Build-Depends: python3-all",
        "",
        "Package: foo",
        "Architecture: all",
    ]
    parse_control = False

    def test_compat(self) -> None:
        with self.assertRaisesRegex(
            Exception, r"Unable to determine debhelper compatibility version."
        ):
            dh = DebHelper(self.build_options(), impl=self.impl)
            dh.compat()



# Copyright © 2012-2013 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging
import os
import re
from os.path import exists, join
from pathlib import Path
from typing import ClassVar, NamedTuple, Self, TYPE_CHECKING

from dhpython import INTERPRETER_DIR_TPLS, PUBLIC_DIR_RE, OLD_SITE_DIRS

if TYPE_CHECKING:
    from typing import Literal


SHEBANG_RE = re.compile(
    r"""
    (?:\#!\s*){0,1}  # shebang prefix
    (?P<path>
        .*?/bin/.*?)?
    (?P<name>
        python|pypy)
    (?P<version>
        \d[\.\d]*)?
    (?P<debug>
        -dbg)?
    (?P<options>.*)
    """,
    re.VERBOSE,
)
EXTFILE_RE = re.compile(
    r"""
    (?P<name>.*?)
    (?:\.
        (?:
            (?P<stableabi>abi\d+)
            |(?P<soabi>
                (?P<impl>cpython|pypy)
                -
                (?P<ver>\d{2,})
                (?P<flags>[a-z]*)
            )
        )
        (?:
            -
            (?P<multiarch>[^/]*?)
        )?
    )?
    (?P<debug>_d)?
    \.so$""",
    re.VERBOSE,
)
log = logging.getLogger("dhpython")


class Shebang(NamedTuple):
    name: str
    path: str | None = None
    version: str | None = None
    debug: bool = False
    options: tuple[str, ...] = ()


class Interpreter:
    """
    :attr path: /usr/bin/ in most cases
    :attr name: pypy or python (even for python3 and python-dbg) or empty string
    :attr version: interpreter's version
    :attr debug: -dbg version of the interpreter
    :attr impl: implementation (cpython3 or pypy)
    :attr options: options parsed from shebang
    :type path: str
    :type name: str
    :type version: Version or None
    :type debug: bool
    :type impl: str
    :type options: tuple
    """

    path = "/usr/bin/"
    name = "python"
    version: "Version | None" = None
    debug = False
    impl = ""
    options: tuple[str, ...] = ()
    _cmd_cache: ClassVar[dict[str, list[str] | str]] = {}
    _re_cache: ClassVar[dict[str, re.Pattern[str]]] = {}

    def __init__(
        self,
        value: Self | str | None = None,
        *,
        path: str | None = None,
        name: str | None = None,
        version: str | None = None,
        debug: str | None = None,
        impl: str | None = None,
        options: str | None = None,
    ) -> None:
        # pylint: disable=unused-argument
        params = locals()
        del params["self"]
        del params["value"]

        if isinstance(value, Interpreter):
            for key in params.keys():
                if params[key] is None:
                    params[key] = getattr(value, key)
        elif value:
            if value.replace(".", "").isdigit() and not version:
                # version string
                params["version"] = Version(value)
            else:
                # shebang or other string
                shebang = self.parse(value)
                assert shebang
                for key, val in shebang._asdict().items():
                    # prefer values passed to constructor over shebang ones:
                    if params[key] is None:
                        params[key] = val

        for key, val in params.items():
            if val is not None:
                setattr(self, key, val)
            elif key == "version":
                setattr(self, key, val)

    def __setattr__(self, name: str, value: "str | Version | None") -> None:
        if name == "name":
            if value not in ("python", "pypy", ""):
                raise ValueError("interpreter not supported: %s" % value)
            if value == "python":
                if self.version:
                    if self.version.major == 3:
                        self.__dict__["impl"] = "cpython3"
            elif value == "pypy":
                if self.version:
                    if self.version.major == 3:
                        self.__dict__["impl"] = "pypy3"
        elif name == "version" and value is not None:
            value = Version(value)
            if not self.impl and self.name == "python":
                if value.major == 3:
                    self.impl = "cpython3"
        if name in ("path", "name", "impl", "options") and value is None:
            pass
        elif name == "debug":
            self.__dict__[name] = bool(value)
        else:
            self.__dict__[name] = value

    def __repr__(self) -> str:
        result = self.path
        if not result.endswith("/"):
            result += "/"
        result += self._vstr(self.version)
        if self.options:
            result += " " + " ".join(self.options)
        return result

    def __str__(self) -> str:
        return self._vstr(self.version)

    def _vstr(
        self, version: "str | Version | None" = None, consider_default_ver: bool = False
    ) -> str:
        if self.impl == "pypy":
            # TODO: will Debian support more than one PyPy version?
            return self.name
        version = version or self.version or ""
        if consider_default_ver and (not version or version == self.default_version):
            version = "3"
        if self.debug:
            return f"python{version}-dbg"
        return self.name + str(version)

    def binary(self, version: "Version | None" = None) -> str:
        return f"{self.path}{self._vstr(version)}"

    @property
    def binary_dv(self) -> str:
        """Like binary(), but returns path to default intepreter symlink
        if version matches default one for given implementation.
        """
        return f"{self.path}{self._vstr(consider_default_ver=True)}"

    @property
    def default_version(self) -> "Version | None":
        if self.impl:
            return default(self.impl)
        return None

    @staticmethod
    def parse(shebang: str) -> Shebang | None:
        """Return dict with parsed shebang

        >>> Interpreter.parse('#!/usr/bin/python')
        Shebang(name='python', path='/usr/bin/', version='3', debug=False, options=())
        >>> Interpreter.parse('/usr/bin/python3.2-dbg')
        Shebang(name='python', path='/usr/bin/', version='3.2', debug=True, options=())
        >>> Interpreter.parse('#! /usr/bin/python3.2')
        Shebang(name='python', path='/usr/bin/', version='3.2', debug=False, options=())
        >>> Interpreter.parse('/usr/bin/python3.2-dbg --foo --bar')
        Shebang(name='python', path='/usr/bin/', version='3.2', debug=True, options=('--foo', '--bar'))
        """
        if not (m := SHEBANG_RE.search(shebang)):
            return None
        groups = m.groupdict()
        version = groups["version"]
        if version is None and groups["name"] == "python":
            version = "3"
        return Shebang(
            name=groups["name"],
            path=groups["path"],
            version=version,
            debug=bool(groups["debug"]),
            # TODO: do we need "--key value" here?
            options=tuple(groups["options"].split()),
        )

    @classmethod
    def from_file(cls, fpath: str | Path) -> "Interpreter":
        """Read file's shebang and parse it."""
        interpreter = Interpreter()
        with open(fpath, "rb") as fp:
            data = fp.read(96)
            if b"\0" in data:
                raise ValueError("cannot parse binary file")
        # make sure only first line is checked
        shebang = data.decode("utf-8").split("\n", maxsplit=1)[0]
        if not shebang.startswith("#!"):
            raise ValueError(f"doesn't look like a shebang: {shebang}")

        parsed = cls.parse(shebang)
        if not parsed:
            raise ValueError(f"doesn't look like a shebang: {shebang}")
        for key, val in parsed._asdict().items():
            setattr(interpreter, key, val)
        return interpreter

    def sitedir(
        self,
        package: str | None = None,
        version: "str | Version | None" = None,
        gdb: bool = False,
    ) -> str:
        """Return path to site-packages directory.

        Note that returned path is not the final location of .py files

        >>> i = Interpreter('python')
        >>> i.sitedir(version='3.1')
        '/usr/lib/python3/dist-packages/'
        >>> i.sitedir(version='3.1', gdb=True, package='python3-foo')
        'debian/python3-foo/usr/lib/debug/usr/lib/python3/dist-packages/'
        >>> i.sitedir(version=Version('3.2'))
        '/usr/lib/python3/dist-packages/'
        """
        try:
            version = Version(version or self.version)
        except Exception as err:
            raise ValueError("cannot find valid version: %s" % err)
        if version << Version("3.0"):
            raise ValueError(f"The version {version} is no longer supported")
        path = "/usr/lib/python3/dist-packages/"

        if gdb:
            path = "/usr/lib/debug%s" % path
        if package:
            path = f"debian/{package}{path}"

        return path

    def old_sitedirs(
        self,
        package: str | None = None,
        version: "str | Version | None" = None,
        gdb: bool = False,
    ) -> list[str]:
        """Return deprecated paths to site-packages directories."""
        try:
            version = Version(version or self.version)
        except Exception as err:
            raise ValueError("cannot find valid version: %s" % err)
        result = []
        for item in OLD_SITE_DIRS.get(self.impl, []):
            if isinstance(item, str):
                result.append(item.format(version))
            else:
                res = item(version)
                if res is not None:
                    result.append(res)

        if gdb:
            result = [f"/usr/lib/debug{i}" for i in result]
            if self.impl.startswith("cpython"):
                result.append(f"/usr/lib/debug/usr/lib/pyshared/python{version}")
        if package:
            result = [f"debian/{package}{i}" for i in result]

        return result

    def parse_public_dir(self, path: str) -> "Version | Literal[True] | None":
        """Return version assigned to site-packages path
        or True is it's unversioned public dir."""
        match = PUBLIC_DIR_RE[self.impl].match(path)
        if match:
            vers = match.group(1)
            if vers and vers[0]:
                return Version(vers)
            return True
        return None

    def should_ignore(self, path: str) -> bool:
        """Return True if path is used by another interpreter implementation."""
        if len(INTERPRETER_DIR_TPLS) == 1:
            return False
        cache_key = f"should_ignore_{self.impl}"
        if cache_key not in self.__class__._re_cache:
            expr = [v for k, v in INTERPRETER_DIR_TPLS.items() if k != self.impl]
            regexp = re.compile("|".join(f"({i})" for i in expr))
            self.__class__._re_cache[cache_key] = regexp
        else:
            regexp = self.__class__._re_cache[cache_key]
        return bool(regexp.search(path))

    @property
    def include_dir(self) -> str:
        """Return INCLUDE_DIR path.

        >>> Interpreter('python3.8').include_dir       # doctest: +SKIP
        '/usr/include/python3.8'
        >>> Interpreter('python3.8-dbg').include_dir   # doctest: +SKIP
        '/usr/include/python3.8d'
        """
        try:
            result = self._get_config()[2]
            if result:
                return result
        except Exception:
            result = ""
            log.debug("cannot get include path", exc_info=True)
        result = f"/usr/include/{self.name}"
        version = self.version
        assert version
        if self.debug:
            if version >= "3.8":
                result += "d"
            elif version << "3.3":
                result += "_d"
            else:
                result += "dm"
        else:
            if version >= "3.8":
                pass
            elif version >> "3.2":
                result += "m"
            elif version == "3.2":
                result += "mu"
        return result

    @property
    def library_file(self) -> str:
        """Return libfoo.so file path."""
        libpl, ldlibrary = self._get_config()[3:5]
        if ldlibrary.endswith(".a"):
            # python3.1-dbg, python3.2, python3.2-dbg returned static lib
            ldlibrary = ldlibrary.replace(".a", ".so")
        if libpl and ldlibrary:
            return join(libpl, ldlibrary)
        raise Exception(f"cannot find library file for {self}")

    def check_extname(self, fname: str, version: "Version | None" = None) -> str | None:
        """Return extension file name if file can be renamed."""
        if not version and not self.version:
            return None

        version = Version(version or self.version)

        if "/" in fname:
            fdir, fname = fname.rsplit("/", 1)  # in case full path was passed
        else:
            fdir = ""

        if not (m := EXTFILE_RE.search(fname)):
            return None
        info = m.groupdict()
        if info["ver"] and (not version or version.minor is None):
            # get version from soabi if version is not set of only major
            # version number is set
            version = Version("{}.{}".format(info["ver"][0], info["ver"][1]))

        if info["stableabi"] and version < Version("3.13"):
            # We added support for stableabi multiarch filenames in 3.13
            # (GH-122931)
            return None
        if info["debug"] and self.debug is False:
            # do not change Python 2.X extensions already marked as debug
            # (the other way around is acceptable)
            return None
        if (info["soabi"] or info["stableabi"]) and info["multiarch"]:
            # already tagged, nothing we can do here
            return None

        try:
            soabi, multiarch = self._get_config(version)[:2]
        except Exception:
            log.debug("cannot get soabi/multiarch", exc_info=True)
            return None

        if info["soabi"] and soabi and info["soabi"] != soabi:
            return None

        tmp_soabi = info["stableabi"] or info["soabi"] or soabi
        tmp_multiarch = info["multiarch"] or multiarch

        result = info["name"]
        if (
            result.endswith("module")
            and result not in ("module", "_module")
            and self.impl == "cpython3"
        ):
            result = result[:-6]

        if tmp_soabi:
            result = f"{result}.{tmp_soabi}"
        if tmp_multiarch and tmp_multiarch not in result:
            result = f"{result}-{tmp_multiarch}"

        result += ".so"
        if fname == result:
            return None
        return join(fdir, result)

    def suggest_pkg_name(self, name: str) -> str:
        """Suggest binary package name with for given library name

        >>> Interpreter('python3.1').suggest_pkg_name('foo')
        'python3-foo'
        >>> Interpreter('python3.8').suggest_pkg_name('foo_bar')
        'python3-foo-bar'
        >>> Interpreter('python3.8-dbg').suggest_pkg_name('bar')
        'python3-bar-dbg'
        """
        name = name.replace("_", "-")
        result = f"python3-{name}"
        if self.debug:
            result += "-dbg"
        return result

    def _get_config(self, version: "str | Version | None" = None) -> list[str]:
        version = Version(version or self.version)
        cmd = (
            'import sysconfig as s; print("__SEP__".join(i or "" '
            "for i in s.get_config_vars("
            '"SOABI", "MULTIARCH", "INCLUDEPY", "LIBPL", "LDLIBRARY")))'
        )
        output = self._execute(cmd, version)
        assert isinstance(output, str)
        conf_vars = output.split("__SEP__")
        if conf_vars[1] in conf_vars[0]:
            # Python >= 3.5 includes MILTIARCH in SOABI
            conf_vars[0] = conf_vars[0].replace("-%s" % conf_vars[1], "")
        try:
            conf_vars[1] = os.environ["DEB_HOST_MULTIARCH"]
        except KeyError:
            pass
        return conf_vars

    def _execute(
        self, command: str, version: "str | Version | None" = None, cache: bool = True
    ) -> list[str] | str:
        version = Version(version or self.version)
        exe = f"{self.path}{self._vstr(version)}"
        command = "{} -c '{}'".format(exe, command.replace("'", "'"))
        if cache and command in self.__class__._cmd_cache:
            return self.__class__._cmd_cache[command]
        if not exists(exe):
            raise Exception(
                "cannot execute command due to missing " "interpreter: %s" % exe
            )

        output = execute(command)
        if output.returncode != 0:
            log.debug(output.stderr)
            raise Exception(f"{command} failed with status code {output.returncode}")

        result: list[str] | str = output.stdout.splitlines()

        if len(result) == 1:
            result = result[0]

        if cache:
            self.__class__._cmd_cache[command] = result

        return result


# due to circular imports issue
from dhpython.tools import execute
from dhpython.version import Version, default



# Copyright © 2025 Stefano Rivera <stefanor@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from dataclasses import asdict, dataclass, field, replace
from typing import Any
from shlex import quote

from dhpython.interpreter import Interpreter
from dhpython.version import Version


@dataclass
class Args:
    ENV: dict[str, str]
    args: str
    autopkgtest: bool
    build_dir: str
    destdir: str
    dir: str
    home_dir: str
    install_dir: str
    interpreter: Interpreter
    package: str
    version: Version
    # during tests only:
    test_dir: str | None = None
    # Injected by base:
    ignore_no_tests: bool = False
    wheel: str | None = None
    # Injected by plugin_distutils:
    setup_py: str | None = None
    parallel: int = 1

    def format(self, template: str) -> str:
        return template.format(**self.as_dict())

    def format_quoted(self, template: str) -> str:
        quoted = self.as_dict()
        for k, v in quoted.items():
            if v and (k in ('dir', 'destdir') or k.endswith('_dir')):
                assert isinstance(v, str)
                quoted[k] = quote(v)
        return template.format(**quoted)

    def as_dict(self) -> dict[str, Any]:
        return asdict(self)


@dataclass
class Context:
    ENV: dict[str, str]
    dir: str
    destdir: str
    args: dict[str, str] = field(default_factory=dict)

    def copy(self) -> "Context":
        return replace(self)



#! /usr/bin/python3
from distutils.core import setup, Extension

setup(name='Foo',
      version='0.1',
      description="package with Python extension",
      author='Piotr Ożarowski',
      author_email='piotr@debian.org',
      url='http://www.debian.org/',
      ext_modules=[Extension('foo/bar', ['lib/bar.c'])],
      #py_modules=['package'],
      packages=['foo'],
      package_dir={'foo': 'lib'})



# Copyright © 2012-2013 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from typing import override

from dhpython.build.base import Base, shell_command, copy_test_files
from dhpython.build.types import Args, Context


class BuildSystem(Base):
    DESCRIPTION = 'CMake build system (using dh_auto_* commands)'
    REQUIRED_COMMANDS = ['cmake']
    REQUIRED_FILES = ['CMakeLists.txt']
    OPTIONAL_FILES = {'cmake_uninstall.cmake': 10, 'CMakeCache.txt': 10}

    @shell_command
    @override
    def clean(self, context: Context, args: Args) -> str:
        super().clean(context, args)
        return 'dh_auto_clean --buildsystem=cmake'

    @shell_command
    @override
    def configure(self, context: Context, args: Args) -> str:
        return ('dh_auto_configure --buildsystem=cmake'
                ' --builddirectory={build_dir} --'
                # FindPythonInterp:
                ' -DPYTHON_EXECUTABLE:FILEPATH=/usr/bin/{interpreter}'
                ' -DPYTHON_LIBRARY:FILEPATH={interpreter.library_file}'
                ' -DPYTHON_INCLUDE_DIR:PATH={interpreter.include_dir}'
                # FindPython:
                ' -DPython_EXECUTABLE=/usr/bin/{interpreter}'
                ' -DPython_LIBRARY={interpreter.library_file}'
                ' -DPython_INCLUDE_DIR={interpreter.include_dir}'
                # FindPython3:
                ' -DPython3_EXECUTABLE=/usr/bin/{interpreter}'
                ' -DPython3_LIBRARY={interpreter.library_file}'
                ' -DPython3_INCLUDE_DIR={interpreter.include_dir}'
                ' {args}')

    @shell_command
    @override
    def build(self, context: Context, args: Args) -> str:
        return ('dh_auto_build --buildsystem=cmake'
                ' --builddirectory={build_dir}'
                ' -- {args}')

    @shell_command
    @override
    def install(self, context: Context, args: Args) -> str:
        return ('dh_auto_install --buildsystem=cmake'
                ' --builddirectory={build_dir}'
                ' --destdir={destdir}'
                ' -- {args}')

    @copy_test_files()
    @shell_command
    @override
    def test(self, context: Context, args: Args) -> str:
        return ('dh_auto_test --buildsystem=cmake'
                ' --builddirectory={build_dir}'
                ' -- {args}')



# Copyright © 2010-2013 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging
import re
from collections.abc import Iterable
from typing import TypeAlias
from os.path import exists

from dhpython import _defaults

RANGE_PATTERN = r"(-)?(\d\.\d+)(?:(-)(\d\.\d+)?)?"
RANGE_RE = re.compile(RANGE_PATTERN)
VERSION_RE = re.compile(
    r"""
    (?P<major>\d+)\.?
    (?P<minor>\d+)?\.?
    (?P<micro>\d+)?[.\s]?
    (?P<releaselevel>alpha|beta|candidate|final)?[.\s]?
    (?P<serial>\d+)?""",
    re.VERBOSE,
)

log = logging.getLogger("dhpython")
Interpreter = None

V: TypeAlias = "str | Version | Iterable[int]"


class Version:
    major: int
    minor: int | None
    micro: int | None
    releaselevel: str | None
    serial: int | None

    # TODO: Upgrade to PEP-440
    def __init__(
        self,
        value: "V | None" = None,
        *,
        major: int | None = None,
        minor: int | None = None,
        micro: int | None = None,
        releaselevel: str | None = None,
        serial: int | None = None,
    ) -> None:
        # pylint: disable=too-many-positional-arguments
        """Construct a new instance.

        >>> Version(major=0, minor=0, micro=0, releaselevel=0, serial=0)
        Version('0.0')
        >>> Version('0.0')
        Version('0.0')
        """
        # pylint: disable=unused-argument
        if isinstance(value, (tuple, list)):
            value = ".".join(str(i) for i in value)
        if isinstance(value, Version):
            for name in ("major", "minor", "micro", "releaselevel", "serial"):
                setattr(self, name, getattr(value, name))
            return
        comp = locals()
        del comp["self"]
        del comp["value"]
        if value:
            assert isinstance(value, str)
            match = VERSION_RE.match(value)
            for name, value in match.groupdict().items() if match else []:
                if value is not None and comp[name] is None:
                    if name == "releaselevel":
                        comp[name] = value
                    else:
                        comp[name] = int(value)
        for name, value in comp.items():
            setattr(self, name, value)
        if self.major is None:
            raise ValueError("major component is required")

    def __str__(self) -> str:
        """Return major.minor or major string.

        >>> str(Version(major=3, minor=2, micro=1, releaselevel='final', serial=4))
        '3.2'
        >>> str(Version(major=2))
        '2'
        """
        result = str(self.major)
        if self.minor is not None:
            result += f".{self.minor}"
        return result

    def __hash__(self) -> int:
        return hash(repr(self))

    def __repr__(self) -> str:
        """Return full version string.

        >>> repr(Version(major=3, minor=2, micro=1, releaselevel='final', serial=4))
        "Version('3.2.1.final.4')"
        >>> repr(Version(major=2))
        "Version('2')"
        """
        result = f"Version('{self}"
        for name in ("micro", "releaselevel", "serial"):
            value = getattr(self, name)
            if not value:
                break
            result += f".{value}"
        return result + "')"

    def __add__(self, other: int | str) -> "Version":
        """Return next version.

        >>> Version('3.1') + 1
        Version('3.2')
        >>> Version('2') + '1'
        Version('3')
        """
        result = Version(self)
        if result.minor is None:
            result.major += int(other)
        else:
            result.minor += int(other)
        return result

    def __sub__(self, other: int | str) -> "Version":
        """Return previous version.

        >>> Version('3.1') - 1
        Version('3.0')
        >>> Version('3') - '1'
        Version('2')
        """
        result = Version(self)
        if result.minor is None:
            result.major -= int(other)
            new = result.major
        else:
            result.minor -= int(other)
            new = result.minor
        if new < 0:
            raise ValueError("cannot decrease version further")
        return result

    def __eq__(self, other: object) -> bool:
        try:
            other = Version(other)  # type: ignore
        except Exception:
            return False
        return self.__cmp(other) == 0

    def __lt__(self, other: object) -> bool:
        return self.__cmp(other) < 0

    def __le__(self, other: object) -> bool:
        return self.__cmp(other) <= 0

    def __gt__(self, other: object) -> bool:
        return self.__cmp(other) > 0

    def __ge__(self, other: object) -> bool:
        return self.__cmp(other) >= 0

    def __lshift__(self, other: V) -> bool:
        """Compare major.minor or major only (if minor is not set).

        >>> Version('2.6') << Version('2.7')
        True
        >>> Version('2.6') << Version('2.6.6')
        False
        >>> Version('3') << Version('2')
        False
        >>> Version('3.1') << Version('2')
        False
        >>> Version('2') << Version('3.2.1.alpha.3')
        True
        """
        if not isinstance(other, Version):
            other = Version(other)
        if self.minor is None or other.minor is None:
            return self.__cmp(other, ignore="minor") < 0
        else:
            return self.__cmp(other, ignore="micro") < 0

    def __rshift__(self, other: V) -> bool:
        """Compare major.minor or major only (if minor is not set).

        >>> Version('2.6') >> Version('2.7')
        False
        >>> Version('2.6.7') >> Version('2.6.6')
        False
        >>> Version('3') >> Version('2')
        True
        >>> Version('3.1') >> Version('2')
        True
        >>> Version('2.1') >> Version('3.2.1.alpha.3')
        False
        """
        if not isinstance(other, Version):
            other = Version(other)
        if self.minor is None or other.minor is None:
            return self.__cmp(other, ignore="minor") > 0
        else:
            return self.__cmp(other, ignore="micro") > 0

    def __cmp(self, other: object, ignore: str | None = None) -> int:
        if not isinstance(other, Version):
            try:
                other = Version(other)  # type: ignore
            except Exception:
                raise ValueError(f"Cannot compare Version with {other!r}")
        for name in ("major", "minor", "micro", "releaselevel", "serial"):
            if name == ignore:
                break
            if name == "releaselevel":
                rmap: dict[str | None, int] = {
                    "alpha": -3,
                    "beta": -2,
                    "candidate": -1,
                    "final": 0,
                }
                value1 = rmap.get(self.releaselevel, 0)
                value2 = rmap.get(other.releaselevel, 0)
            else:
                value1 = getattr(self, name) or 0
                value2 = getattr(other, name) or 0
            if value1 == value2:
                continue
            return (value1 > value2) - (value1 < value2)
        return 0


class VersionRange:
    minver: Version | None
    maxver: Version | None

    def __init__(
        self,
        value: str | None = None,
        minver: "V | None" = None,
        maxver: "V | None" = None,
    ) -> None:
        if minver:
            self.minver = Version(minver)
        else:
            self.minver = None
        if maxver:
            self.maxver = Version(maxver)
        else:
            self.maxver = None

        if value:
            minver, maxver = self.parse(value)
            if minver and self.minver is None:
                self.minver = minver
            if maxver and self.maxver is None:
                self.maxver = maxver

    def __bool__(self) -> bool:
        if self.minver is not None or self.maxver is not None:
            return True
        return False

    def __str__(self) -> str:
        """Return version range string from given range.

        >>> str(VersionRange(minver='3.4'))
        '3.4-'
        >>> str(VersionRange(minver='3.4', maxver='3.6'))
        '3.4-3.6'
        >>> str(VersionRange(minver='3.4', maxver='4.0'))
        '3.4-4.0'
        >>> str(VersionRange(maxver='3.7'))
        '-3.7'
        >>> str(VersionRange(minver='3.5', maxver='3.5'))
        '3.5'
        >>> str(VersionRange())
        '-'
        """
        if self.minver is None is self.maxver:
            return "-"
        if self.minver == self.maxver:
            return str(self.minver)
        elif self.minver is None:
            return f"-{self.maxver}"
        elif self.maxver is None:
            return f"{self.minver}-"
        else:
            return f"{self.minver}-{self.maxver}"

    def __repr__(self) -> str:
        """Return version range string.

        >>> repr(VersionRange('5.0-'))
        "VersionRange(minver='5.0')"
        >>> repr(VersionRange('3.0-3.5'))
        "VersionRange(minver='3.0', maxver='3.5')"
        """
        result = "VersionRange("
        if self.minver is not None:
            result += f"minver='{self.minver}'"
            if self.maxver is not None:
                result += ", "
        if self.maxver is not None:
            result += f"maxver='{self.maxver}'"
        return result + ")"

    @staticmethod
    def parse(value: str) -> tuple[Version | None, Version | None]:
        """Return minimum and maximum Python version from given range.

        >>> VersionRange.parse('3.0-')
        (Version('3.0'), None)
        >>> VersionRange.parse('3.1-3.13')
        (Version('3.1'), Version('3.13'))
        >>> VersionRange.parse('3.2-4.0')
        (Version('3.2'), Version('4.0'))
        >>> VersionRange.parse('-3.7')
        (None, Version('3.7'))
        >>> VersionRange.parse('3.2')
        (Version('3.2'), Version('3.2'))
        >>> VersionRange.parse('') == VersionRange.parse('-')
        True
        >>> VersionRange.parse('>= 4.0')
        (Version('4.0'), None)
        """
        if value in ("", "-"):
            return None, None

        match = RANGE_RE.match(value)
        if not match:
            try:
                minv, maxv = VersionRange._parse_pycentral(value)
            except Exception:
                raise ValueError("version range is invalid: %s" % value)
        else:
            groups = match.groups()

            if list(groups).count(None) == 3:  # only one version is allowed
                minv = Version(groups[1])
                return minv, minv

            minv_s: str | None = None
            maxv_s: str | None = None
            if groups[0]:  # maximum version only
                maxv_s = groups[1]
            else:
                minv_s = groups[1]
                maxv_s = groups[3]

            minv = Version(minv_s) if minv_s else None
            maxv = Version(maxv_s) if maxv_s else None

        if maxv and minv and minv > maxv:
            raise ValueError("version range is invalid: %s" % value)

        return minv, maxv

    @staticmethod
    def _parse_pycentral(value: str) -> tuple[Version | None, Version | None]:
        """Parse X-Python3-Version.

        >>> VersionRange._parse_pycentral('>= 3.10')
        (Version('3.10'), None)
        >>> VersionRange._parse_pycentral('<< 4.0')
        (None, Version('4.0'))
        >>> VersionRange._parse_pycentral('3.1')
        (Version('3.1'), Version('3.1'))
        >>> VersionRange._parse_pycentral('3.1, 3.2')
        (Version('3.1'), None)
        """

        minv = maxv = None
        hardcoded = set()

        for item in value.split(","):
            item = item.strip()

            match = re.match(r">=\s*([\d\.]+)", item)
            if match:
                minv = match.group(1)
                continue
            match = re.match(r"<<\s*([\d\.]+)", item)
            if match:
                maxv = match.group(1)
                continue
            match = re.match(r"^[\d\.]+$", item)
            if match:
                hardcoded.add(match.group(0))

        if len(hardcoded) == 1:
            ver = hardcoded.pop()
            return Version(ver), Version(ver)

        if not minv and hardcoded:
            # yeah, no maxv!
            minv = sorted(hardcoded)[0]

        return Version(minv) if minv else None, Version(maxv) if maxv else None


def default(impl: str) -> Version:
    """Return default interpreter version for given implementation."""
    if impl not in _defaults.DEFAULT:
        raise ValueError("interpreter implementation not supported: %r" % impl)
    ver = _defaults.DEFAULT[impl]
    return Version(major=ver[0], minor=ver[1])


def supported(impl: str) -> list[Version]:
    """Return list of supported interpreter versions for given implementation."""
    if impl not in _defaults.SUPPORTED:
        raise ValueError("interpreter implementation not supported: %r" % impl)
    versions = _defaults.SUPPORTED[impl]
    return [Version(major=v[0], minor=v[1]) for v in versions]


def get_requested_versions(
    impl: str,
    vrange: VersionRange | str | None = None,
    available: bool | None = None,
) -> set[Version]:
    """Return a set of requested and supported Python versions.

    :param impl: interpreter implementation
    :param available: if set to `True`, return installed versions only,
        if set to `False`, return requested versions that are not installed.
        By default returns all requested versions.
    :type available: bool

    >>> sorted(get_requested_versions('cpython3', '')) == sorted(supported('cpython3'))
    True
    >>> sorted(get_requested_versions('cpython3', '-')) == sorted(supported('cpython3'))
    True
    >>> get_requested_versions('cpython3', '>= 5.0')
    set()
    """
    from dhpython.interpreter import Interpreter

    parsed_vrange: VersionRange | None = None
    if isinstance(vrange, str):
        parsed_vrange = VersionRange(vrange)

    if not parsed_vrange:
        versions = set(supported(impl))
    else:
        minv = (
            Version(major=0, minor=0)
            if parsed_vrange.minver is None
            else parsed_vrange.minver
        )
        maxv = (
            Version(major=99, minor=99)
            if parsed_vrange.maxver is None
            else parsed_vrange.maxver
        )
        if minv == maxv:
            versions = set([minv] if minv in supported(impl) else tuple())
        else:
            versions = {v for v in supported(impl) if minv <= v < maxv}

    if available:
        interpreter = Interpreter(impl=impl)
        versions = {v for v in versions if exists(interpreter.binary(v))}
    elif available is False:
        interpreter = Interpreter(impl=impl)
        versions = {v for v in versions if not exists(interpreter.binary(v))}

    return versions


def build_sorted(
    versions: Iterable[V],
    impl: str = "cpython3",
) -> list[Version]:
    """Return sorted list of versions in a build friendly order.

    i.e. default version, if among versions, is sorted last.

    >>> build_sorted([(2, 6), (3, 4), default('cpython3'), (3, 6), (2, 7)])[-1] == default('cpython3')
    True
    >>> build_sorted(('3.2', (3, 0), '3.1'))
    [Version('3.0'), Version('3.1'), Version('3.2')]
    """
    default_ver = default(impl)

    result = sorted(Version(v) for v in versions)
    try:
        result.remove(default_ver)
    except ValueError:
        pass
    else:
        result.append(default_ver)
    return result



import sys
import unittest
from os import environ
from os.path import exists
from dhpython.interpreter import Interpreter


class TestInterpreter(unittest.TestCase):
    def setUp(self) -> None:
        self._triplet = environ.get("DEB_HOST_MULTIARCH")
        environ["DEB_HOST_MULTIARCH"] = "MYARCH"

    def tearDown(self) -> None:
        if self._triplet:
            environ["DEB_HOST_MULTIARCH"] = self._triplet
        else:
            del environ["DEB_HOST_MULTIARCH"]

    @unittest.skipUnless(exists("/usr/bin/python3.11"), "python3.11 is not installed")
    def test_python311(self) -> None:
        i = Interpreter("python3.11")
        self.assertEqual(i.check_extname("foo.so"), r"foo.cpython-311-MYARCH.so")
        self.assertIsNone(i.check_extname("foo.cpython-32m.so"))  # different version
        self.assertIsNone(
            i.check_extname("foo.cpython-311-OTHER.so")
        )  # different architecture
        self.assertEqual(
            i.check_extname("foo.cpython-311.so"), r"foo.cpython-311-MYARCH.so"
        )
        self.assertIsNone(i.check_extname("foo.abi3.so"))
        self.assertEqual(
            i.check_extname("foo/bar/bazmodule.so"),
            r"foo/bar/baz.cpython-311-MYARCH.so",
        )

    @unittest.skipUnless(
        exists("/usr/bin/python3.11-dbg"), "python3.11-dbg is not installed"
    )
    def test_python311dbg(self) -> None:
        i = Interpreter("python3.11-dbg")
        self.assertEqual(i.check_extname("foo.so"), r"foo.cpython-311d-MYARCH.so")
        self.assertIsNone(i.check_extname("foo.cpython-32m.so"))  # different version
        self.assertIsNone(
            i.check_extname("foo.cpython-311-OTHER.so")
        )  # different architecture
        self.assertIsNone(i.check_extname("foo.abi3.so"))
        self.assertEqual(
            i.check_extname("foo/bar/bazmodule.so"),
            r"foo/bar/baz.cpython-311d-MYARCH.so",
        )

    @unittest.skipUnless(exists("/usr/bin/python3.12"), "python3.12 is not installed")
    def test_python312(self) -> None:
        i = Interpreter("python3.12")
        self.assertEqual(i.check_extname("foo.so"), r"foo.cpython-312-MYARCH.so")
        self.assertIsNone(i.check_extname("foo.cpython-32m.so"))  # different version
        self.assertIsNone(
            i.check_extname("foo.cpython-312-OTHER.so")
        )  # different architecture
        self.assertEqual(
            i.check_extname("foo.cpython-312.so"), r"foo.cpython-312-MYARCH.so"
        )
        self.assertIsNone(i.check_extname("foo.abi3.so"))
        self.assertEqual(
            i.check_extname("foo/bar/bazmodule.so"),
            r"foo/bar/baz.cpython-312-MYARCH.so",
        )

    @unittest.skipUnless(
        exists("/usr/bin/python3.12-dbg"), "python3.12-dbg is not installed"
    )
    def test_python312dbg(self) -> None:
        i = Interpreter("python3.12-dbg")
        self.assertEqual(i.check_extname("foo.so"), r"foo.cpython-312d-MYARCH.so")
        self.assertIsNone(i.check_extname("foo.cpython-32m.so"))  # different version
        self.assertIsNone(
            i.check_extname("foo.cpython-312-OTHER.so")
        )  # different architecture
        self.assertIsNone(i.check_extname("foo.abi3.so"))
        self.assertEqual(
            i.check_extname("foo/bar/bazmodule.so"),
            r"foo/bar/baz.cpython-312d-MYARCH.so",
        )

    @unittest.skipUnless(exists("/usr/bin/python3.13"), "python3.13 is not installed")
    def test_python313(self) -> None:
        i = Interpreter("python3.13")
        self.assertEqual(i.check_extname("foo.so"), r"foo.cpython-313-MYARCH.so")
        self.assertIsNone(i.check_extname("foo.cpython-33m.so"))  # different version
        self.assertIsNone(
            i.check_extname("foo.cpython-313-OTHER.so")
        )  # different architecture
        self.assertEqual(
            i.check_extname("foo.cpython-313.so"), r"foo.cpython-313-MYARCH.so"
        )
        self.assertEqual(i.check_extname("foo.abi3.so"), "foo.abi3-MYARCH.so")
        self.assertIsNone(i.check_extname("foo.abi3-OTHER.so"))
        self.assertEqual(
            i.check_extname("foo/bar/bazmodule.so"),
            r"foo/bar/baz.cpython-313-MYARCH.so",
        )

    @unittest.skipUnless(
        exists("/usr/bin/python3.13-dbg"), "python3.13-dbg is not installed"
    )
    def test_python313dbg(self) -> None:
        i = Interpreter("python3.13-dbg")
        self.assertEqual(i.check_extname("foo.so"), r"foo.cpython-313d-MYARCH.so")
        self.assertIsNone(i.check_extname("foo.cpython-33m.so"))  # different version
        self.assertIsNone(
            i.check_extname("foo.cpython-313-OTHER.so")
        )  # different architecture
        self.assertEqual(i.check_extname("foo.abi3.so"), "foo.abi3-MYARCH.so")
        self.assertIsNone(i.check_extname("foo.abi3-OTHER.so"))
        self.assertEqual(
            i.check_extname("foo/bar/bazmodule.so"),
            r"foo/bar/baz.cpython-313d-MYARCH.so",
        )

    @unittest.skipUnless(exists("/usr/bin/python3.14"), "python3.14 is not installed")
    def test_python314(self) -> None:
        i = Interpreter("python3.14")
        self.assertEqual(i.check_extname("foo.so"), r"foo.cpython-314-MYARCH.so")
        self.assertIsNone(i.check_extname("foo.cpython-34m.so"))  # different version
        self.assertIsNone(
            i.check_extname("foo.cpython-314-OTHER.so")
        )  # different architecture
        self.assertEqual(
            i.check_extname("foo.cpython-314.so"), r"foo.cpython-314-MYARCH.so"
        )
        self.assertEqual(i.check_extname("foo.abi3.so"), "foo.abi3-MYARCH.so")
        self.assertIsNone(i.check_extname("foo.abi3-OTHER.so"))
        self.assertEqual(
            i.check_extname("foo/bar/bazmodule.so"),
            r"foo/bar/baz.cpython-314-MYARCH.so",
        )

    @unittest.skipUnless(
        exists("/usr/bin/python3.14-dbg"), "python3.14-dbg is not installed"
    )
    def test_python314dbg(self) -> None:
        i = Interpreter("python3.14-dbg")
        self.assertEqual(i.check_extname("foo.so"), r"foo.cpython-314d-MYARCH.so")
        self.assertIsNone(i.check_extname("foo.cpython-34m.so"))  # different version
        self.assertIsNone(
            i.check_extname("foo.cpython-314-OTHER.so")
        )  # different architecture
        self.assertEqual(i.check_extname("foo.abi3.so"), "foo.abi3-MYARCH.so")
        self.assertIsNone(i.check_extname("foo.abi3-OTHER.so"))
        self.assertEqual(
            i.check_extname("foo/bar/bazmodule.so"),
            r"foo/bar/baz.cpython-314d-MYARCH.so",
        )

    def test_python3(self) -> None:
        i = Interpreter("python{}.{}".format(*sys.version_info[:2]))
        pyver = "{}{}".format(*sys.version_info[:2])
        self.assertEqual(i.check_extname("foo.so"), rf"foo.cpython-{pyver}-MYARCH.so")
        self.assertIsNone(i.check_extname("foo.cpython-32m.so"))  # different version
        self.assertIsNone(
            i.check_extname(f"foo.cpython-{pyver}-OTHER.so")
        )  # different architecture
        self.assertEqual(
            i.check_extname(f"foo.cpython-{pyver}.so"),
            rf"foo.cpython-{pyver}-MYARCH.so",
        )
        self.assertEqual(i.check_extname("foo.abi3.so"), "foo.abi3-MYARCH.so")
        self.assertIsNone(i.check_extname("foo.abi3-OTHER.so"))
        self.assertEqual(
            i.check_extname("foo/bar/bazmodule.so"),
            rf"foo/bar/baz.cpython-{pyver}-MYARCH.so",
        )

    @unittest.skipUnless(exists("/usr/bin/python3-dbg"), "python3-dbg is not installed")
    def test_python3dbg(self) -> None:
        i = Interpreter("python{}.{}-dbg".format(*sys.version_info[:2]))
        pyver = "{}{}".format(*sys.version_info[:2])
        self.assertEqual(i.check_extname("foo.so"), rf"foo.cpython-{pyver}d-MYARCH.so")
        self.assertIsNone(i.check_extname("foo.cpython-32m.so"))  # different version
        self.assertIsNone(
            i.check_extname(f"foo.cpython-{pyver}-OTHER.so")
        )  # different architecture
        self.assertEqual(i.check_extname("foo.abi3.so"), "foo.abi3-MYARCH.so")
        self.assertIsNone(i.check_extname("foo.abi3-OTHER.so"))
        self.assertEqual(
            i.check_extname("foo/bar/bazmodule.so"),
            rf"foo/bar/baz.cpython-{pyver}d-MYARCH.so",
        )

    def test_bare_module(self) -> None:
        i = Interpreter("python{}.{}".format(*sys.version_info[:2]))
        pyver = "{}{}".format(*sys.version_info[:2])
        self.assertEqual(
            i.check_extname("module.so"), rf"module.cpython-{pyver}-MYARCH.so"
        )
        self.assertEqual(
            i.check_extname("_module.so"), rf"_module.cpython-{pyver}-MYARCH.so"
        )


if __name__ == "__main__":
    unittest.main()



# Copyright © 2010-2013 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import logging
import re
from dataclasses import dataclass
from os import makedirs, chmod, environ
from os.path import basename, exists, join, dirname
from sys import argv
from typing import NamedTuple, TypeAlias

from dhpython import DEPENDS_SUBSTVARS, PKG_NAME_TPLS, RT_LOCATIONS, RT_TPLS

log = logging.getLogger("dhpython")
parse_dep = re.compile(
    r"""[,\s]*
    (?P<name>[^\s:]+)(?::any)?
    \s*
    \(?(?P<version>([>=<]{2,}|=)\s*[^\)]+)?\)?
    \s*
    (?:\[(?P<arch>[^\]]+)\])?
    """,
    re.VERBOSE,
).match


@dataclass
class PackageData:
    substvars: dict[str, list[str]]
    autoscripts: dict[str, dict[str, list[str]]]
    rtupdates: list[tuple[str, str]]
    arch: str


class Options(NamedTuple):
    arch: bool | None
    package: list[str]
    no_package: list[str]
    write_log: bool
    compile_all: bool
    remaining_packages: bool


def build_options(
    *,
    arch: bool | None = None,
    package: list[str] | None = None,
    no_package: list[str] | None = None,
    write_log: bool = False,
    compile_all: bool = False,
    remaining_packages: bool = False,
) -> Options:
    return Options(
        arch=arch,
        compile_all=compile_all,
        package=package or [],
        no_package=no_package or [],
        write_log=write_log,
        remaining_packages=remaining_packages,
    )


BD: TypeAlias = dict[str, dict[str | None, str]]


class DebHelper:
    """Reinvents the wheel / some dh functionality (Perl is ugly ;-P)"""

    options: Options
    packages: dict[str, PackageData]
    build_depends: BD

    def __init__(self, options: Options, impl: str = "cpython3") -> None:
        self.options = options
        self.packages = {}
        self.build_depends = {}
        self.python_version = None
        self.impl = impl
        self.command = {
            "cpython3": "dh_python3",
        }[impl]
        skip_tpl_set: set[str] = set()
        for name, tpls in PKG_NAME_TPLS.items():
            if name != impl:
                skip_tpl_set.update(tpls)
        skip_tpl = tuple(skip_tpl_set)
        substvar = DEPENDS_SUBSTVARS[impl]

        pkgs = options.package
        skip_pkgs = options.no_package

        try:
            with open("debian/control", encoding="utf-8") as fp:
                paragraphs: list[dict[str, str]] = [{}]
                field = None
                for lineno, line in enumerate(fp, 1):
                    if line.startswith("#"):
                        continue
                    if not line.strip():
                        if paragraphs[-1]:
                            paragraphs.append({})
                            field = None
                        continue
                    if line[0].isspace():  # Continuation
                        assert field
                        paragraphs[-1][field] += line.rstrip()
                        continue
                    if not ":" in line:
                        raise Exception(
                            "Unable to parse line %i in debian/control: %s"
                            % (lineno, line)
                        )
                    field, value = line.split(":", 1)
                    field = field.lower()
                    paragraphs[-1][field] = value.strip()
        except OSError as e:
            if e.errno == errno.ENOENT:
                raise Exception("cannot find debian/control file")
            raise

        # Trailing new lines?
        if not paragraphs[-1]:
            paragraphs.pop()

        if len(paragraphs) < 2:
            raise Exception(
                "Unable to parse debian/control, found less than 2 paragraphs"
            )

        self.source_name = paragraphs[0]["source"]
        if self.impl == "cpython3" and "x-python3-version" in paragraphs[0]:
            self.python_version = paragraphs[0]["x-python3-version"]
            if len(self.python_version.split(",")) > 2:
                raise ValueError(
                    "too many arguments provided for "
                    "X-Python3-Version: min and max only."
                )

        build_depends_list = []
        for field in ("build-depends", "build-depends-indep", "build-depends-arch"):
            if field in paragraphs[0]:
                build_depends_list.append(paragraphs[0][field])
        build_depends = ", ".join(build_depends_list)
        for dep1 in build_depends.split(","):
            for dep2 in dep1.split("|"):
                if m := parse_dep(dep2):
                    details = m.groupdict()
                    if details["arch"]:
                        architectures = details["arch"].split()
                    else:
                        architectures = [None]
                    for arch in architectures:
                        self.build_depends.setdefault(details["name"], {})[arch] = (
                            details["version"]
                        )

        for paragraph_no, paragraph in enumerate(paragraphs[1:], 2):
            if "package" not in paragraph:
                raise Exception(
                    "Unable to parse debian/control, paragraph %i "
                    "missing Package field" % paragraph_no
                )
            binary_package = paragraph["package"]
            if skip_tpl and binary_package.startswith(skip_tpl):
                log.debug("skipping package: %s", binary_package)
                continue
            if pkgs and binary_package not in pkgs:
                continue
            if skip_pkgs and binary_package in skip_pkgs:
                continue
            if options.remaining_packages and self.has_acted_on_package(binary_package):
                continue
            pkg = PackageData(
                substvars={},
                autoscripts={},
                rtupdates=[],
                arch=paragraph["architecture"],
            )
            if (
                options.arch is False
                and pkg.arch != "all"
                or options.arch is True
                and pkg.arch == "all"
            ):
                # TODO: check also if arch matches current architecture:
                continue

            if (
                not binary_package.startswith(PKG_NAME_TPLS[impl])
                and self.compat() < 14
            ):
                # package doesn't have common prefix (python3-)
                # so lets check if Depends/Recommends contains the
                # appropriate substvar
                if substvar not in paragraph.get(
                    "depends", ""
                ) and substvar not in paragraph.get("recommends", ""):
                    log.debug(
                        "skipping package %s (missing %s in Depends/Recommends)",
                        binary_package,
                        substvar,
                    )
                    continue
            # Operate on binary_package
            self.packages[binary_package] = pkg

        fp.close()
        log.debug(
            "source=%s, binary packages=%s",
            self.source_name,
            list(self.packages.keys()),
        )

    def compat(self) -> int:
        if "debhelper-compat" in self.build_depends:
            version = self.build_depends["debhelper-compat"][None]
            assert version.startswith("=")
            assert version[1:].strip().isdigit()
            return int(version[1:].strip())

        if exists("debian/compat"):
            with open("debian/compat", encoding="UTF-8") as f:
                compat = int(f.read())
            if compat >= 15:
                raise Exception(
                    f"debian/compat is no longer supported for level {compat}."
                )
            return compat
        raise Exception("Unable to determine debhelper compatibility version.")

    def has_acted_on_package(self, package: str) -> bool:
        try:
            with open(f"debian/{package}.debhelper.log", encoding="utf-8") as f:
                for line in f:
                    if line.strip() == self.command:
                        return True
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
        return False

    def addsubstvar(self, package: str, name: str, value: str) -> None:
        """debhelper's addsubstvar"""
        self.packages[package].substvars.setdefault(name, []).append(value)

    def autoscript(self, package: str, when: str, template: str, args: str) -> None:
        """debhelper's autoscript"""
        self.packages[package].autoscripts.setdefault(when, {}).setdefault(
            template, []
        ).append(args)

    def add_rtupdate(self, package: str, value: tuple[str, str]) -> None:
        self.packages[package].rtupdates.append(value)

    def save_autoscripts(self) -> None:
        for package, settings in self.packages.items():
            autoscripts = settings.autoscripts
            if not autoscripts:
                continue

            for when, templates in autoscripts.items():
                fn = f"debian/{package}.{when}.debhelper"
                if exists(fn):
                    with open(fn, encoding="utf-8") as datafile:
                        data = datafile.read()
                else:
                    data = ""

                new_data = ""
                for tpl_name, args in templates.items():
                    for i in args:
                        # try local one first (useful while testing dh_python3)
                        fpath = join(
                            dirname(__file__), "..", "autoscripts/%s" % tpl_name
                        )
                        if not exists(fpath):
                            fpath = "/usr/share/debhelper/autoscripts/%s" % tpl_name
                        with open(fpath, encoding="utf-8") as tplfile:
                            tpl = tplfile.read()
                        if self.options.compile_all and args:
                            # TODO: should args be checked to contain dir name?
                            tpl = tpl.replace("-p #PACKAGE#", "")
                        elif settings.arch == "all":
                            tpl = tpl.replace("#PACKAGE#", package)
                        else:
                            arch = environ["DEB_HOST_ARCH"]
                            tpl = tpl.replace("#PACKAGE#", f"{package}:{arch}")
                        tpl = tpl.replace("#ARGS#", i)
                        if tpl not in data and tpl not in new_data:
                            new_data += "\n%s" % tpl
                if new_data:
                    data += (
                        f"\n# Automatically added by {basename(argv[0])}"
                        + f"{new_data}\n# End automatically added section\n"
                    )
                    with open(fn, "w", encoding="utf-8") as fp:
                        fp.write(data)

    def save_substvars(self) -> None:
        for package, settings in self.packages.items():
            substvars = settings.substvars
            if not substvars:
                continue
            fn = "debian/%s.substvars" % package
            if exists(fn):
                with open(fn, encoding="utf-8") as datafile:
                    data = datafile.read()
            else:
                data = ""
            for name, values in substvars.items():
                p = data.find("%s=" % name)
                if p > -1:  # parse the line and remove it from data
                    e = data[p:].find("\n")
                    line = data[p + len("%s=" % name) : p + e if e > -1 else None]
                    items = [i.strip() for i in line.split(",") if i]
                    if e > -1 and data[p + e :].strip():
                        data = f"{data[:p]}\n{data[p + e :]}"
                    else:
                        data = data[:p]
                else:
                    items = []
                for j in values:
                    if j not in items:
                        items.append(j)
                if items:
                    if data:
                        data += "\n"
                    data += "{}={}\n".format(name, ", ".join(items))
            data = data.replace("\n\n", "\n")
            if data:
                with open(fn, "w", encoding="utf-8") as fp:
                    fp.write(data)

    def save_rtupdate(self) -> None:
        for package, settings in self.packages.items():
            pkg_arg = "" if self.options.compile_all else "-p %s" % package
            values = settings.rtupdates
            if not values:
                continue
            d = f"debian/{package}/{RT_LOCATIONS[self.impl]}"
            if not exists(d):
                makedirs(d)
            fn = f"{d}/{package}.rtupdate"
            if exists(fn):
                with open(fn, encoding="utf-8") as fp:
                    data = fp.read()
            else:
                data = "#! /bin/sh\nset -e"
            for dname, args in values:
                cmd = RT_TPLS[self.impl].format(pkg_arg=pkg_arg, dname=dname, args=args)
                if cmd not in data:
                    data += "\n%s" % cmd
            if data:
                with open(fn, "w", encoding="utf-8") as fp:
                    fp.write(data)
                chmod(fn, 0o755)

    def save_log(self) -> None:
        if not self.options.write_log:
            return
        for package, _ in self.packages.items():
            with open(f"debian/{package}.debhelper.log", "a", encoding="utf-8") as f:
                f.write(self.command + "\n")

    def save(self) -> None:
        self.save_substvars()
        self.save_autoscripts()
        self.save_rtupdate()
        self.save_log()



# Copyright © 2012-2013 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging
from collections.abc import Callable
from functools import wraps
from glob import glob1
from os import environ, remove, walk, makedirs
from os.path import exists, isdir, join
from pathlib import Path
from shutil import rmtree, copyfile, copytree, which
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Concatenate,
    Literal,
    ParamSpec,
    TypeVar,
    cast,
    overload,
)

from dhpython import CACHE_DIRECTORIES
from dhpython.debhelper import DebHelper, build_options
from dhpython.exceptions import RequiredCommandMissingException
from dhpython.tools import execute, ExecutionResult, ExecutionResultWithOutput
from dhpython.build.options import PybuildOptions
from dhpython.build.types import Args, Context

if TYPE_CHECKING:
    from functools import _Wrapped

log = logging.getLogger('dhpython')

P = ParamSpec('P')
T = TypeVar('T')
BC = TypeVar('BC', bound="Base")

def copy_test_files(
    dest: str = '{build_dir}',
    filelist: str = '{home_dir}/testfiles_to_rm_before_install',
    add_to_args: tuple[str, ...] = ('test', 'tests')
) -> Callable[
        [Callable[Concatenate[BC, Context, Args, P], T]],
        Callable[Concatenate[BC, Context, Args, P], T],
    ]:

    def _copy_test_files(
        func: Callable[Concatenate[BC, Context, Args, P], T], /
    ) -> "_Wrapped[Concatenate[BC, Context, Args, P], T, Concatenate[BC, Context, Args, P], T]":

        @wraps(func)
        def __copy_test_files(
            self: BC,
            context: Context,
            args: Args,
            *oargs: P.args,
            **kwargs: P.kwargs
        ) -> T:
            files_to_copy = {'pyproject.toml', 'pytest.ini', 'test', 'tests'}
            # check debian/pybuild_pythonX.Y.testfiles
            for tpl in ('_{i}{v}', '_{i}{m}', ''):
                tpl = tpl.format(i=args.interpreter.name,
                                 v=args.version,
                                 m=args.version.major)
                fpath = join(args.dir, f'debian/pybuild{tpl}.testfiles')
                if exists(fpath):
                    with open(fpath, encoding='utf-8') as fp:
                        # overwrite files_to_copy if .testfiles file found
                        files_to_copy = set()
                        for line in fp.readlines():
                            line = line.strip()
                            if line.startswith('#'):
                                continue
                            if not line:
                                continue
                            files_to_copy.add(line)
                        break

            files_to_remove = set()
            for entry in files_to_copy:
                name, *rest = entry.split(maxsplit=1)
                dest_suffix = rest[0] if rest else ""
                if not dest_suffix or dest_suffix.endswith('/'):
                    dest_suffix = join(dest_suffix, name.rsplit('/', 1)[-1])

                src_dpath = join(args.dir, name)
                dst_dpath = join(args.format(dest), dest_suffix)
                if exists(src_dpath):
                    if not exists(dst_dpath):
                        log.debug("Copying %s to %s for tests", src_dpath, dst_dpath)
                        makedirs(dst_dpath.rsplit('/', 1)[0], exist_ok=True)
                        if isdir(src_dpath):
                            copytree(src_dpath, dst_dpath)
                        else:
                            copyfile(src_dpath, dst_dpath)
                        files_to_remove.add(dst_dpath + '\n')
                    if not args.args and 'PYBUILD_TEST_ARGS' not in context.ENV\
                       and (self.cfg.test_pytest or self.cfg.test_nose) \
                       and name in add_to_args:
                        args.args = name
            if files_to_remove and filelist:
                with open(args.format(filelist), 'a', encoding="UTF-8") as fp:
                    fp.writelines(files_to_remove)

            return func(self, context, args, *oargs, **kwargs)
        return __copy_test_files
    return _copy_test_files


class Base:
    """Base class for build system plugins

    :attr REQUIRED_COMMANDS: list of command checked by default in :meth:is_usable,
        if one of them is missing, plugin cannot be used.
    :type REQUIRED_COMMANDS: list of strings
    :attr REQUIRED_FILES: list of files (or glob templates) required by given
        build system
    :attr OPTIONAL_FILES: dictionary of glob templates (key) and score (value)
        used to detect if given plugin is the best one for the job
    :type OPTIONAL_FILES: dict (key is a string, value is an int)
    :attr SUPPORTED_INTERPRETERS: set of interpreter templates (with or without
        {version}) supported by given plugin
    """
    cfg: PybuildOptions
    NAME: ClassVar[str]
    DESCRIPTION = ''
    REQUIRED_COMMANDS: list[str] = []
    REQUIRED_FILES: list[str] = []
    DETECTED_REQUIRED_FILES: dict[str, list[str]]
    OPTIONAL_FILES: dict[str, int] = {}
    DETECTED_OPTIONAL_FILES: dict[str, list[str]]
    SUPPORTED_INTERPRETERS: Literal[True] | set[str] = {
        'python', 'python3', 'python-dbg', 'python3-dbg',
        'python{version}', 'python{version}-dbg'
    }
    # files and directories to remove during clean step (other than .pyc):
    CLEAN_FILES = CACHE_DIRECTORIES

    def __init__(self, cfg: PybuildOptions) -> None:
        self.cfg = cfg

    def __repr__(self) -> str:
        return "BuildSystem(%s)" % self.NAME

    @classmethod
    def is_usable(cls) -> None:
        for command in cls.REQUIRED_COMMANDS:
            pth = which(command)
            if not pth:
                raise RequiredCommandMissingException(command)

    def detect(self, context: Context) -> int:
        """Return certainty level that this plugin describes the right build system

        This method is using cls.{REQUIRED,OPTIONAL}_FILES only by default,
        please extend it in the plugin if more sofisticated methods can be used
        for given build system.

        :return: 0 <= certainty <= 100
        :rtype: int
        """
        result = 0

        required_files_num = 0
        self.DETECTED_REQUIRED_FILES = {}  # can be used in the plugin later
        for tpl in self.REQUIRED_FILES:
            found = False
            for ftpl in tpl.split('|'):
                res = glob1(context.dir, ftpl)
                if res:
                    found = True
                    self.DETECTED_REQUIRED_FILES.setdefault(tpl, []).extend(res)
            if found:
                required_files_num += 1
        # add max 50 points depending on how many required files are available
        if self.REQUIRED_FILES:
            result += int(required_files_num / len(self.REQUIRED_FILES) * 50)

        self.DETECTED_OPTIONAL_FILES = {}
        for ftpl, score in self.OPTIONAL_FILES.items():
            res = glob1(context.dir, ftpl)
            if res:
                result += score
                self.DETECTED_OPTIONAL_FILES.setdefault(ftpl, []).extend(res)
        if result > 100:
            return 100
        return result

    def clean(self, context: Context, args: Args) -> None:
        tox_dir = join(args.dir, '.tox')
        if isdir(tox_dir):
            try:
                rmtree(tox_dir)
            except Exception:
                log.debug('cannot remove %s', tox_dir)

        for fn in self.CLEAN_FILES:
            path = join(context.dir, fn)
            if isdir(path):
                try:
                    rmtree(path)
                except Exception:
                    log.debug('cannot remove %s', path)
            elif exists(path):
                try:
                    remove(path)
                except Exception:
                    log.debug('cannot remove %s', path)

        dh = DebHelper(build_options())
        # Plugins that rely on repository contents to build MANIFEST
        clean_sources_txt = not {
            'python3-setuptools-scm', 'python3-setuptools-git'
        }.intersection(set(dh.build_depends))

        for root, dirs, file_names in walk(context.dir):
            for name in dirs[:]:
                if name == '__pycache__' or (
                        clean_sources_txt and name.endswith('.egg-info')):
                    dpath = join(root, name)
                    log.debug('removing dir: %s', dpath)
                    try:
                        rmtree(dpath)
                    except Exception:
                        log.debug('cannot remove %s', dpath)
                    else:
                        dirs.remove(name)
            for fn in file_names:
                if fn.endswith(('.pyc', '.pyo')):
                    fpath = join(root, fn)
                    log.debug('removing: %s', fpath)
                    try:
                        remove(fpath)
                    except Exception:
                        log.debug('cannot remove %s', fpath)
            if root.endswith('.egg-info'):
                for fn in file_names:
                    if fn != 'SOURCES.txt':
                        fpath = join(root, fn)
                        log.debug('removing: %s', fpath)
                        try:
                            remove(fpath)
                        except Exception:
                            log.debug('cannot remove %s', fpath)

    def configure(self, context: Context, args: Args) -> None:
        raise NotImplementedError("configure method not implemented in %s" % self.NAME)

    def install(self, context: Context, args: Args) -> None:
        raise NotImplementedError("install method not implemented in %s" % self.NAME)

    def build(self, context: Context, args: Args) -> None:
        raise NotImplementedError("build method not implemented in %s" % self.NAME)

    def test(self, context: Context, args: Args) -> None:
        raise NotImplementedError("test method not implemented in %s" % self.NAME)

    @copy_test_files()
    def test_cmd(self, context: Context, args: Args) -> str:
        if self.cfg.test_nose2:
            return 'cd {build_dir}; {interpreter} -m nose2 -v {args}'
        elif self.cfg.test_nose:
            return 'cd {build_dir}; {interpreter} -m nose -v {args}'
        elif self.cfg.test_pytest:
            return 'cd {build_dir}; {interpreter} -m pytest {args}'
        elif self.cfg.test_tox:
            tox_config = None
            if exists(join(args.dir, 'tox.ini')):
                tox_config = '{dir}/tox.ini'
            elif exists(join(args.dir, 'pyproject.toml')):
                tox_config = '{dir}/pyproject.toml'
            elif exists(join(args.dir, 'setup.cfg')):
                tox_config = '{dir}/setup.cfg'
            else:
                raise Exception("tox config not found. "
                    "Expected to find tox.ini, pyproject.toml, or setup.cfg")

            tox_cmd = ['cd {build_dir};',
                   'tox',
                   '-c', tox_config,
                   '--sitepackages',
                   '-e', 'py{version.major}{version.minor}',
                   '-x', 'testenv.passenv+=_PYTHON_HOST_PLATFORM',
            ]
            if args.autopkgtest:
                tox_cmd += ['--skip-pkg-install']

            if not args.autopkgtest:
                wheel = self.built_wheel(context, args)
                if not wheel:
                    self.build_wheel(context, args)
                    wheel = self.built_wheel(context, args)
                assert wheel
                args.wheel = wheel
                tox_cmd += ['--installpkg', '{wheel}']

            tox_cmd.append('{args}')
            return ' '.join(tox_cmd)
        elif self.cfg.test_stestr:
            return (
                'cd {build_dir};'
                'stestr --config {dir}/.stestr.conf init;'
                'PYTHON=python{version} stestr --config {dir}/.stestr.conf run'
            )
        elif self.cfg.test_unittest:
            return 'cd {build_dir}; {interpreter} -m unittest discover -v {args}'
        elif self.cfg.test_custom:
            return 'cd {build_dir}; {args}'
        else:
            log.warning(
                "No test runner selected, defaulting to unittest, ignoring "
                "test discovery problems.\nUse --test-unittest to explicitly "
                "select the unittest runner.")
            # As this is the fallback option, we allow 0 discovered tests.
            args.ignore_no_tests = True
            return 'cd {build_dir}; {interpreter} -m unittest discover -v {args}'

    def build_wheel(self, context: Context, args: Args) -> None:
        raise NotImplementedError("build_wheel method not implemented in %s" % self.NAME)

    def built_wheel(self, context: Context, args: Args) -> str | None:
        """Return the path to any built wheels we can find"""
        # pylint: disable=unused-argument
        wheels = list(Path(args.home_dir).glob('*.whl'))
        n_wheels = len(wheels)
        if n_wheels > 1:
            raise Exception(f"Expecting to have built exactly 1 wheel, but found {n_wheels}")
        if n_wheels == 1:
            return str(wheels[0])
        return None

    @overload
    def execute(
        self,
        context: Context,
        args: Args,
        command: str,
        *,
        log_file: None = None,
    ) -> ExecutionResultWithOutput:
        ...

    @overload
    def execute(
        self,
        context: Context,
        args: Args,
        command: str,
        *,
        log_file: str | Literal[False],
    ) -> ExecutionResult:
        ...

    def execute(
        self,
        context: Context,
        args: Args,
        command: str,
        *,
        log_file: str | Literal[False] | None = None
    ) -> ExecutionResult | ExecutionResultWithOutput:
        if log_file is False and self.cfg.really_quiet:
            log_file = None
        command = args.format(command)
        env = dict(context.ENV)
        env.update(args.ENV)
        log.info(command)
        return execute(command, cwd=context.dir, env=env, log_output=log_file)

    def print_args(self, context: Context, args: Args) -> None:
        # pylint: disable=unused-argument
        cfg = self.cfg
        if len(cfg.print_args) == 1 and len(cfg.interpreter) == 1 and '{version}' not in cfg.interpreter[0]:
            i = cfg.print_args[0]
            if '{' in i:
                print(args.format(i))
            else:
                print(args.as_dict().get(i, ''))
        else:
            for i in cfg.print_args:
                if '{' in i:
                    print(args.format(i))
                else:
                    print('{} {}: {}'.format(args.interpreter, i, args.as_dict().get(i, '')))


def shell_command(
    func: Callable[Concatenate[BC, Context, Args, P], str | Literal[0]], /
) -> Callable[Concatenate[BC, Context, Args, P], None]:

    @wraps(func)
    def wrapped_func(
        self: BC,
        context: Context,
        args: Args,
        *oargs: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        command = kwargs.pop('command', None)
        if not command:
            command = func(self, context, args, *oargs, **kwargs)
            if isinstance(command, int):  # final result
                return
        assert isinstance(command, str)
        if not command:
            log.warning('missing command '
                     '(plugin=%s, method=%s, interpreter=%s, version=%s)',
                     self.NAME, func.__name__,
                     args.interpreter, args.version)
            return

        log_file: str | Literal[False]
        if self.cfg.quiet:
            log_file = join(args.home_dir, f'{func.__name__}_cmd.log')
        else:
            log_file = False

        command = args.format_quoted(command)

        output = self.execute(context, args, command=command, log_file=log_file)
        if output.returncode == 5 and args.ignore_no_tests:
            # Temporary hack (see Base.test)
            pass
        elif output.returncode != 0:
            msg = f'exit code={output.returncode}: {command}'
            if log_file:
                msg += f'\nfull command log is available in {log_file}'
            raise Exception(msg)

    # Workaround https://github.com/python/typeshed/issues/10653
    return cast(Callable[Concatenate[BC, Context, Args, P], None], wrapped_func)


def create_setuptools_parallel_cfg(
    backend: Callable[[], str | None] | None = None
) -> Callable[
        [Callable[Concatenate[BC, Context, Args, P], T]],
        Callable[Concatenate[BC, Context, Args, P], T],
]:
    """Enable parallel building of extensions

    An adaption of Michał Górny's technique described in
    https://blogs.gentoo.org/mgorny/2024/03/15/optimizing-parallel-extension-builds-in-pep517-builds

    This function is not part of :pyfunc:`create_pydistutils_cfg` as that
    function is not used for `build_wheel`, which is our target.

    For PEP-517 (pyproject) builds we can not yet use PEP-517 Config Settings
    as we need to pass the parallel building option to `build_ext` and
    setuptools (as of version 80.9.0) does not have a way to redirect PEP-517
    Config Settings to specific commands. Setuptools only passes PEP-517 Config
    Settings to all commands via `--global-option` and to `build_wheel` with
    `--build-option`.
    See https://github.com/pypa/setuptools/blob/d198e86f57231e83de87975c5c82bc40c196da79/setuptools/build_meta.py
    and https://github.com/pypa/setuptools/discussions/4083
    """
    def decorator(
        func: Callable[Concatenate[BC, Context, Args, P], T], /
    ) -> Callable[Concatenate[BC, Context, Args, P], T]:
        @wraps(func)
        def wrapped_func(
            self: BC,
            context: Context,
            args: Args,
            *oargs: P.args,
            **kwargs: P.kwargs,
        ) -> T:
            fpath = join(args.home_dir, 'setuptools-build_ext-parallel.cfg')
            if not exists(fpath):
                with open(fpath, 'w', encoding='utf-8') as fp:
                    lines = ['[build_ext]\n',
                             f'parallel={args.parallel}\n']
                    log.debug('parallel build extension config file:\n%s', ''.join(lines))
                    fp.writelines(lines)
            context.ENV['DIST_EXTRA_CONFIG'] = fpath
            return func(self, context, args, *oargs, **kwargs)

        if 'DIST_EXTRA_CONFIG' in environ:
            return func

        if backend is not None:
            pyproject_backend = backend()
            if pyproject_backend is not None and pyproject_backend != "setuptools.build_meta":
                return func

        return cast(Callable[Concatenate[BC, Context, Args, P], T], wrapped_func)
    return decorator



#! /usr/bin/python3
# Copyright © 2013 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging
from configparser import ConfigParser
from os import environ
from os.path import exists
from subprocess import Popen, PIPE
from typing import NamedTuple

SUPPORTED: dict[str, tuple[tuple[int, int], ...]] = {
    "cpython3": ((3, 8),),
}
DEFAULT: dict[str, tuple[int, int]] = {
    "cpython3": (3, 8),
}

log = logging.getLogger("dhpython")


class CPythonVersions(NamedTuple):
    default: tuple[int, int] | None
    supported: tuple[tuple[int, int], ...]


def _parse_version(version: str) -> tuple[int, int]:
    parts = version.split(".")
    assert len(parts) == 2
    return (int(parts[0]), int(parts[1]))


def cpython_versions(major: int) -> CPythonVersions:
    assert major > 2
    default_ver: tuple[int, int] | None = None
    supported_vers: tuple[tuple[int, int], ...] = ()
    ver = str(major)
    supported = environ.get(f"DEBPYTHON{ver}_SUPPORTED")
    default = environ.get(f"DEBPYTHON{ver}_DEFAULT")
    if not supported or not default:
        config = ConfigParser()
        config.read(f"/usr/share/python{ver}/debian_defaults")
        if not default:
            default = config.get("DEFAULT", "default-version", fallback="")[6:]
        if not supported:
            supported = config.get(
                "DEFAULT", "supported-versions", fallback=""
            ).replace("python", "")
    if default:
        try:
            default_ver = _parse_version(default)
        except Exception as err:
            log.warning("invalid debian_defaults file: %s", err)
    if supported:
        try:
            supported_vers = tuple(_parse_version(i) for i in supported.split(","))
        except Exception as err:
            log.warning("invalid debian_defaults file: %s", err)
    return CPythonVersions(
        default=default_ver,
        supported=supported_vers,
    )


def from_file(fpath: str) -> None:
    if not exists(fpath):
        raise ValueError("missing interpreter: %s" % fpath)
    command = f"{fpath} --version"
    with Popen(command, shell=True, stdout=PIPE, encoding="utf-8") as process:
        stdout, _ = process.communicate()

    print(stdout)


cpython3 = cpython_versions(3)
if cpython3.default:
    DEFAULT["cpython3"] = cpython3.default
if cpython3.supported:
    SUPPORTED["cpython3"] = cpython3.supported


if __name__ == "__main__":
    import sys

    if len(sys.argv) != 3:
        print("invalid number of arguments", file=sys.stderr)
        sys.exit(1)
    if sys.argv[1] == "default":
        print(".".join(str(i) for i in DEFAULT[sys.argv[2]]))
    elif sys.argv[1] == "supported":
        print(",".join(".".join(str(i) for i in v) for v in SUPPORTED[sys.argv[2]]))



# Copyright © 2010-2013 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import re

PKG_PREFIX_MAP = {"cpython3": "python3"}

# minimum version required for compile/clean scripts:
MINPYCDEP = {"cpython3": "python3:any"}

PUBLIC_DIR_RE = {
    "cpython3": re.compile(r".*?/usr/lib/python(3(?:\.\d+)?)(?:/|$)"),
}

INTERPRETER_DIR_TPLS = {
    "cpython3": r".*/python3(?:\.\d+)?/",
}

MULTIARCH_DIR_TPL = re.compile(
    ".*/([a-z][^/-]+-(?:linux|kfreebsd|gnu)(?:-[^/-]+)?)(?:/.*|$)"
)

# Interpreter site-directories
OLD_SITE_DIRS = {
    "cpython3": [
        "/usr/local/lib/python{}/site-packages",
        "/usr/local/lib/python{}/dist-packages",
        "/usr/lib/python{}/site-packages",
        "/usr/lib/python{}/dist-packages",
        "/var/lib/python-support/python{}",
        "/usr/lib/pymodules/python{}",
    ],
}

# PyDist related
PYDIST_DIRS = {
    "cpython3": "/usr/share/python3/dist/",
}

PYDIST_OVERRIDES_FNAMES = {
    "cpython3": "debian/py3dist-overrides",
}

PYDIST_DPKG_SEARCH_TPLS = {
    # implementation: (dpkg -S query, regex filter)
    "cpython3": ("*python3/*/{}-?*.*-info", r".(egg|dist)-info$"),
}

# DebHelper related
DEPENDS_SUBSTVARS = {
    "cpython3": "${python3:Depends}",
}
PKG_NAME_TPLS = {
    "cpython3": ("python3-", "python3."),
}
RT_LOCATIONS = {
    "cpython3": "/usr/share/python3/runtime.d/",
}
RT_TPLS = {
    "cpython3": """
if [ "$1" = rtupdate ]; then
\tpy3clean {pkg_arg} {dname}
\tpy3compile {pkg_arg} {args} {dname} || true
fi""",
}

# Delete in dh_python3 and pybuild clean
CACHE_DIRECTORIES = frozenset({".coverage", ".pdm-build", ".pytest_cache"})
